import os
import logging
import ConfigParser
from collections import OrderedDict
from scipy.interpolate import interp1d
from bisect import bisect_left
import numpy as np
from itertools import tee, izip

import json
import matplotlib.pyplot as plt
import matplotlib.ticker as plticker
import matplotlib.gridspec as gridspec
from matplotlib import rcParams
#import pandas as pd
# SEABORN creates by default plots with a filled background!!
#import seaborn as sns
from astropy.io import ascii
from astropy.io import fits

import set_shared_labels as shLab
import sys
import dependencies.WeightedKDE as WeightedKDE
import dependencies.autoscale as autoscale
from dependencies.walker_random_sampling import WalkerRandomSampling
from dependencies import FillBetweenStep

from beagle_utils import BeagleDirectories, prepare_plot_saving, set_plot_ticks, plot_exists
from beagle_filters import PhotometricFilters
from beagle_summary_catalogue import BeagleSummaryCatalogue
#from beagle_residual_photometry import ResidualPhotometry
from beagle_multinest_catalogue import MultiNestCatalogue
from beagle_posterior_predictive_checks import PosteriorPredictiveChecks
from beagle_mock_catalogue import BeagleMockCatalogue

# See here
# http://peak.telecommunity.com/DevCenter/PythonEggs#accessing-package-resources
# for an explanation on this approach to include data files
from pkg_resources import resource_stream

TOKEN_SEP = ":"
microJy = np.float32(1.E-23 * 1.E-06)
nanoJy = np.float32(1.E-23 * 1.E-09)

p_value_lim = 0.05


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return izip(a, b)

class ObservedSpectrum(object):

    def __init__(self):

        self.description  = None

    def configure(self, param_file=None, config=None):

        if param_file is None:
            param_file = os.path.join(BeagleDirectories.results_dir, BeagleDirectories.param_file)

        if config is None:
            config = ConfigParser.SafeConfigParser()
            config.read(param_file)

        line = config.get('main', 'SPECTRUM FILE DESCRIPTION')

        self.description = { 
                "wl"        : {"colName" : None, "conversion" : None, "dispersion" : None, "type" : None}, 
                "flux"      : {"colName" : None, "conversion" : None}, 
                "fluxerr"   : {"colName" : None},  
                "sky"       : {"colName" : None},  
                "mask"      : {"colName" : None},  
                "redshift"  : {"keyword" : "redshift"},  
                "min_rel_err" : None,  
                }

        for key, value in self.description.iteritems():

            token = key + TOKEN_SEP

            if isinstance(value, dict):
                for in_key, in_value in value.iteritems():
                    in_token = token + in_key + TOKEN_SEP
                    if in_token in line:
                        self.description[key][in_key] = line.split(in_token)[1].split(" ")[0]

            else:
                if token in line:
                    self.description[key] = line.split(token)[1].split(" ")[0]
                

    def load(self, file_name):

        """ 
        Load an observed spectrum. It automatically
        detects, and loads, FITS or ASCII files depending on the suffix.

        Parameters
        ----------
        file_name : str
            Contains the file name of the spectrum.
        """

        if self.description is None:
            msg = ("The `ObservedSpectrum.description` must be set through "
                    "the `configure` method!")
            raise AttributeError(msg)


        hdu = fits.open(os.path.expandvars(file_name))
        data = hdu[1].data

        self.data = dict()

        # Set the array containing the wavelength bins of the spectrum 
        try:
            self.data['wl'] = np.array(data[self.description["wl"]["colName"]])
        except:
            msg = ("`" + self.description["wl"]["colName"] + "` not found "
                    "in the current spectrum!")
            raise AttributeError(msg)

        if self.description["wl"]["conversion"] is not None:
            self.data['wl'] *= np.float(self.description["wl"]["conversion"])

        # Set the array containing the flux
        try:
            self.data['flux'] = np.array(data[self.description["flux"]["colName"]])
        except:
            msg = ("`" + self.description["flux"]["colName"] + "` not found "
                    "in the current spectrum!")
            raise AttributeError(msg)

        if self.description["flux"]["conversion"] is not None:
            self.data['flux'] *= np.float(self.description["flux"]["conversion"])

        # Set the array containing the flux error
        if self.description["fluxerr"]["colName"] is not None:
            try:
                self.data['fluxerr'] = np.array(data[self.description["fluxerr"]["colName"]])
            except:
                msg = ("`" + self.description["fluxerr"]["colName"] + "` not found "
                        "in the current spectrum!")
                raise AttributeError(msg)

        # Set the redshift
        if self.description["redshift"]["keyword"] is not None:
            self.data['redshift'] = np.float(hdu[1].header[self.description["redshift"]["keyword"]])

        # Set the mask spectrum
        if self.description["mask"]["colName"] is not None:
            self.data['mask'] = np.array(data[self.description["mask"]["colName"]], dtype=bool)


        hdu.close()

class Spectrum(object):

    def __init__(self, params_file,
            line_labels_json=None,
            plot_line_labels=False,
            resolution=None,
            mock_catalogue=None,
            wl_range=None,
            wl_units='micron',
            plot_full_SED=False,
            print_ID=False,
            wl_rest=False,
            log_flux=False,
            draw_steps=False,
            show_residual=False,
            plot_suffix=None,
            n_SED_to_plot=100):

        self.observed_spectrum = ObservedSpectrum()

        self.multinest_catalogue = MultiNestCatalogue()

        self.mock_catalogue = mock_catalogue

        #self.residual = ResidualPhotometry()

        self.PPC = PosteriorPredictiveChecks()

        if line_labels_json == None:
            self.line_labels = json.load(resource_stream(__name__, 'files/emission_lines.json'), 
                    object_pairs_hook=OrderedDict)
        else:
            with open(line_labels_json) as f:    
                 self.line_labels = json.load(f, object_pairs_hook=OrderedDict)

        self.plot_line_labels = plot_line_labels

        self.resolution = resolution
    
        self.wl_range = wl_range

        self.wl_units = wl_units

        self.wl_rest = wl_rest

        self.log_flux = log_flux

        self.plot_full_SED = plot_full_SED

        self.show_residual = show_residual

        self.print_ID = print_ID

        self.draw_steps = draw_steps

        self.n_SED_to_plot = n_SED_to_plot

        self.plot_suffix = plot_suffix

    def plot_marginal(self, ID, 
            observation_name=None,
            max_interval=95.0,
            print_text=False, 
            replot=False):    
        """ 
        Plot the fluxes predicted by BEAGLE.

        The fluxes here considered are those predicted by BEAGLE, given the
        posterior distribution of the model parameters. These are *not*
        replicated data.

        Parameters
        ----------
        ID : int
            ID of the galaxy whose marginal photometry will be plotted.

        max_interval : float, optional
            The marginal photometry is shown to include `max_interval`
            probability, e.g. `max_interval` = 68. will show the 68 % (i.e.
            '1-sigma') (central) credible region of the marginal photometry.

        print_text : bool, optional
            Whether to print further information on the plot, such as
            chi-square, p-value, or leave it empty and neat.

        print_text : bool, optional
            Whether to print the object ID on the top of the plot.
        """

        # Factor to convert angstrom to input units
        if self.wl_units == 'micron':
            wl_factor = 1.E+04
            xlabel = "\mu\\textnormal{m}"
        elif self.wl_units == 'nm':
            wl_factor = 1.E+01
            xlabel = "\\textnormal{nm}"
        elif self.wl_units == 'ang':
            wl_factor = 1.
            xlabel = "\\textnormal{\\AA}"
        else:
            raise ValueError("Wavelength units `" + self.wl_units + "` not recognised!")

        # If needed load the observed spectrum
        if observation_name is not None:
            self.observed_spectrum.load(observation_name)
    
        # Name of the output plot
        plot_suffix = '' 
        if self.plot_suffix is not None: plot_suffix = '_' + self.plot_suffix
        if self.wl_rest:
            plot_name = str(ID) + '_BEAGLE_marginal_SED_spec_rest_wl' + plot_suffix + '.pdf'
        else:
            plot_name = str(ID) + '_BEAGLE_marginal_SED_spec' + plot_suffix + '.pdf'

        # Check if the plot already exists
        if plot_exists(plot_name) and not replot:
            logging.warning('The plot "' + plot_name + '" already exists. \n Exiting the function.')
            return

        # The observed spectrum
        observation = self.observed_spectrum

        # Add to the error array the minimum relative error thet BEAGLE allows
        # one to add to the errors quoted in the catalogue
        #for i, err in enumerate(self.filters.data['flux_errcolName']):
        #    tmp_err = observation[0][err]
        #    if tmp_err > 0.:
        #        obs_flux_err[i] = observation[0][err]*aper_corr*self.filters.units / nanoJy
        #        obs_flux_err[i] = (np.sqrt( (obs_flux_err[i]/obs_flux[i])**2 +
        #                np.float32(self.filters.data['min_rel_err'][i])**2) *
        #                obs_flux[i])
        #    else:
        #        obs_flux_err[i] = tmp_err

        #ok = np.where(obs_flux_err > 0.)[0]

        # Open the file containing BEAGLE results
        fits_file = os.path.join(BeagleDirectories.results_dir,
                str(ID) + '_' + BeagleDirectories.suffix + '.fits.gz')

        hdulist = fits.open(fits_file)

        # Read the template wl array, and the 2D flux array
        model_wl = hdulist['marginal sed wl'].data['wl'][0,:]
        model_fluxes = hdulist['marginal sed'].data

        # Read the posterior probability
        probability = hdulist['posterior pdf'].data['probability']

        # Now compute for each wl bin the sorted fluxes, you will need this to
        # calculate the median and percentiles for each wl bin
        sort_indices = np.argsort(model_fluxes, axis=0)

        # Now it's time to compute the median (observed-frame) SED and its percentiles
        n_wl = model_fluxes.shape[1]

        median_flux = np.zeros(n_wl)
        lower_flux = np.zeros(n_wl)
        upper_flux = np.zeros(n_wl)

        for i in range(n_wl):

            # Compute the cumulative probability
            # ******************************************************************
            # Here you must simply use `cumsum`, and not `cumtrapz` as in
            # beagle_utils.prepare_violin_plot, since the output of MultiNest are a set
            # of weights (which sum up to 1) associated to each set of parameters (the
            # `p_j` of equation 9 of Feroz+2009), and not a probability density (as the
            # MultiNest README would suggest).
            # ******************************************************************
            sort_ = sort_indices[:,i]
            cumul_pdf = np.cumsum(probability[sort_])
            cumul_pdf /= cumul_pdf[len(cumul_pdf)-1]
            
            # Get the interpolant of the cumulative probability
            f_interp = interp1d(cumul_pdf, model_fluxes[sort_,i])

            # The median corresponds to a cumulative probability = 0.5
            median_flux[i] = f_interp(0.5)

            # Compute the percentiles for the different credible regions
            lev = (1.-max_interval/100.)/2.
            lower_flux[i] = f_interp(lev)

            lev = 1.-(1.-max_interval/100.)/2.
            upper_flux[i] = f_interp(lev)
    
        # Set the plot limits from the minimum and maximum wl_eff
        axs = list()
        residual_axs = list()
        z1 = (1.+observation.data['redshift'])
        if self.wl_rest:
            data_wl = observation.data['wl'] / z1
            data_flux = observation.data['flux'] * z1
            data_flux_err = observation.data['fluxerr'] * z1

            model_wl /= z1
            median_flux *= z1
            lower_flux *= z1
            upper_flux *= z1
        else:
            data_wl = observation.data['wl']
            data_flux = observation.data['flux']
            data_flux_err = observation.data['fluxerr']

        # Convert to correct wl units
        model_wl /= wl_factor
        data_wl /= wl_factor

        # Load the mask spectrum if exists
        mask_color = "grey"
        has_mask = False
        data_mask = np.ones(len(data_wl), dtype=bool)
        if "mask" in observation.data:
            data_mask = observation.data['mask']
            has_mask = True

        # Create masked versions of arrays
        slices = list()
        masked_array = np.ma.array(data_wl, mask=~data_mask)
        slices.append(np.ma.clump_unmasked(masked_array))
        slices.append(np.ma.clump_masked(masked_array))

        if self.show_residual:
            n_outer = 2
        else:
            n_outer = 1

        if self.wl_range is None:
            n_ranges = 1
        else:
            if len(self.wl_range) == 2:
                n_ranges = 1
            else:
                n_ranges = int(1.*len(self.wl_range)/2.)

        fig = plt.figure(figsize=(12,8))
        fig, axs_ = plt.subplots(n_outer, n_ranges, gridspec_kw = {'height_ratios':[3, 1]})
        fig.subplots_adjust(wspace=0.1, hspace=0.0)
        axs = axs_[0,:] 
        if self.show_residual:
            residual_axs = axs_[1,:]

        if self.wl_range is None:
            dwl = data_wl[-1]-data_wl[0]
            wl_low = data_wl[0] - dwl*0.025
            wl_up = data_wl[-1] + dwl*0.025
            axs[0].set_xlim([wl_low, wl_up])
            if self.show_residual:
                residual_axs[0].set_xlim([wl_low, wl_up])
        else:
            if len(self.wl_range) == 2:
                axs[0].set_xlim(self.wl_range)
                if self.show_residual:
                    residual_axs[0].set_xlim(self.wl_range)
            else:
                # how big to make the diagonal lines in axes coordinates
                # converting "points" to axes coordinates: 
                # https://stackoverflow.com/a/33638091
                t = axs[0].transAxes.transform([(0,0), (1,1)])
                t = axs[0].get_figure().get_dpi() / (t[1,1] - t[0,1]) / 72
                d = 0.5*(rcParams['xtick.major.size']*t)

                wl_ranges = [(self.wl_range[2*i], self.wl_range[2*i+1]) for i in range(n_ranges)]

                wl_l = wl_ranges[0]
                for i, (ax_l, ax_r) in enumerate(pairwise(axs)):
                    kwargs = dict(transform=ax_l.transAxes, color='k', clip_on=False)
                    ax_l.spines['right'].set_visible(False)
                    if not ax_l.spines['left'].get_visible():
                        ax_l.yaxis.set_ticks_position('none')
                    else:
                        ax_l.yaxis.tick_left()
                    ax_l.set_xlim(wl_l)
                    ax_l.plot((1-d, 1+d), (-d, +d), **kwargs)        
                    ax_l.plot((1-d, 1+d), (1-d, 1+d), **kwargs)        

                    kwargs = dict(transform=ax_r.transAxes, color='k', clip_on=False)
                    ax_r.spines['left'].set_visible(False)
                    ax_r.yaxis.tick_right()
                    ax_r.tick_params(labelright='off') 
                    wl_r = wl_ranges[1+i]
                    ax_r.set_xlim(wl_r)
                    ax_r.plot((-d, +d), (-d, +d), **kwargs)        
                    ax_r.plot((-d, +d), (1-d, 1+d), **kwargs)        

                    wl_l = wl_r

                if self.show_residual:
                    t = residual_axs[0].transAxes.transform([(0,0), (1,1)])
                    t = residual_axs[0].get_figure().get_dpi() / (t[1,1] - t[0,1]) / 72
                    d = 0.5*(rcParams['xtick.major.size']*t)

                    wl_l = wl_ranges[0]
                    for i, (ax_l, ax_r) in enumerate(pairwise(residual_axs)):
                        kwargs = dict(transform=ax_l.transAxes, color='k', clip_on=False)
                        ax_l.spines['right'].set_visible(False)
                        if not ax_l.spines['left'].get_visible():
                            ax_l.yaxis.set_ticks_position('none')
                        else:
                            ax_l.yaxis.tick_left()
                        ax_l.set_xlim(wl_l)
                        ax_l.plot((1-d, 1+d), (-d, +d), **kwargs)        
                        #ax_l.plot((1-d, 1+d), (1-d, 1+d), **kwargs)        

                        kwargs = dict(transform=ax_r.transAxes, color='k', clip_on=False)
                        ax_r.spines['left'].set_visible(False)
                        ax_r.yaxis.tick_right()
                        ax_r.tick_params(labelright='off') 
                        wl_r = wl_ranges[1+i]
                        ax_r.set_xlim(wl_r)
                        ax_r.plot((-d, +d), (-d, +d), **kwargs)        
                        #ax_r.plot((-d, +d), (1-d, 1+d), **kwargs)        
                        #ax_l.spines['top'].set_visible(False)
                        ax_l.spines['top'].set_color('none')
                        ax_l.xaxis.set_ticks_position('bottom')

                        ax_r.spines['top'].set_color('none')
                        ax_r.xaxis.set_ticks_position('bottom')
                        #ax_r.spines['top'].set_visible(False)

                        ax_r.patch.set_facecolor('None')
                        ax_l.patch.set_facecolor('None')

                        wl_l = wl_r

        which = 'both'
        for ax in axs:
            if self.log_flux:
                ax.set_yscale('log')
                which = 'x'

            # Set better location of tick marks
            if self.wl_range is not None:
                set_plot_ticks(ax, which=which, prune_x='both', n_x=3)
                for tick in ax.get_xticklabels():
                    tick.set_rotation(45)
                    tick.set_ha('left')
                    tick.set_rotation_mode("anchor")
            else:
                set_plot_ticks(ax, which=which)

        if self.show_residual:
            for ax in axs:
                ax.tick_params(labelbottom='off')

            for ax in residual_axs:
                if self.wl_range is not None:
                    set_plot_ticks(ax, prune_x='both', n_x=3)
                    for tick in ax.get_xticklabels():
                        tick.set_rotation(45)

        # Define plotting styles
        xlabel = "$\uplambda / " + xlabel + "$"
        if not self.wl_rest:
            xlabel = xlabel + "(observed frame)"

        if self.wl_range is not None:
            fig.text(0.5, -0.02, xlabel, ha='center')
        else:
            fig.text(0.5, 0.02, xlabel, ha='center')


        ylabel = "$F_{\\uplambda} / (\\textnormal{erg} \, \
                \\textnormal{s}^{-1} \, \\textnormal{cm}^{-2} \, \
                \\textnormal{\AA}^{-1})$"
        axs[0].set_ylabel(ylabel)

        if self.show_residual:
            ylabel = "$(F_{\\uplambda}-F_{\\uplambda}^\\textnormal{mod}) / F_{\\uplambda}$"
            residual_axs[0].set_ylabel(ylabel)

        # Title of the plot is the object ID
        if self.print_ID: 
            #fig.text(0.5, 0.95, str(ID).split('_')[0].strip(), ha='center', va='top')
            plt.suptitle(str(ID).split('_')[0].strip())

        colors = ["red", mask_color]

        for ax in axs:

            for slic, col in zip(slices, colors):
                for s in slic:
                    if (self.draw_steps):
                        ax.step(data_wl[s],
                                data_flux[s],
                                where="mid",
                                color=col,
                                linewidth=1.50,
                                alpha=0.7
                                )
                    else:
                        ax.plot(data_wl[s],
                                data_flux[s],
                                color=col,
                                linewidth=2.00,
                                alpha=0.7
                                )

                    if (self.draw_steps):
                        FillBetweenStep.fill_between_steps(ax,
                                data_wl[s],
                                data_flux[s]-data_flux_err[s],
                                data_flux[s]+data_flux_err[s],
                                step_where="mid",
                                color=col, 
                                linewidth=0,
                                interpolate=True,
                                alpha=0.3
                                )
                    else:
                        ax.fill_between(data_wl[s],
                                data_flux[s]-data_flux_err[s],
                                data_flux[s]+data_flux_err[s],
                                facecolor=col, 
                                linewidth=0,
                                interpolate=True,
                                alpha=0.3
                                )


            kwargs = { 'alpha': 0.7 }
            if (self.draw_steps):
                ax.step(model_wl,
                        median_flux,
                        where="mid",
                        color = "blue",
                        linewidth = 1.0,
                        **kwargs
                        )
            else:
                ax.plot(model_wl,
                        median_flux,
                        color = "blue",
                        linewidth = 1.5,
                        **kwargs
                        )

            kwargs = { 'alpha': 0.3 }
            if (self.draw_steps):
                FillBetweenStep.fill_between_steps(ax,
                        model_wl,
                        lower_flux[:], 
                        upper_flux[:],
                        step_where="mid",
                        color = "blue", 
                        linewidth=0,
                        **kwargs)
            else:
                ax.fill_between(model_wl,
                        lower_flux[:],
                        upper_flux[:],
                        facecolor = "blue", 
                        linewidth=0,
                        interpolate=True,
                        **kwargs)

            # Extract and plot full SED
            if 'full sed wl' in hdulist and self.plot_full_SED:
                indices = np.arange(len(probability))
                wrand = WalkerRandomSampling(probability, keys=indices)
                rand_indices = wrand.random(self.n_SED_to_plot)

                if self.wl_rest:
                    wl_obs = hdulist['full sed wl'].data['wl'][0,:]

                    for i in rand_indices:
                        SED = hdulist['full sed'].data[i,:]
                else:
                    z1 = (1.+self.observed_spectrum.data['redshift'])
                    wl_obs = hdulist['full sed wl'].data['wl'][0,:] * z1

                    for i in rand_indices:
                        SED = hdulist['full sed'].data[i,:] / z1


                    ax.plot(wl_obs, 
                            flux_obs,
                            color="black",
                            ls="-",
                            lw=0.5,
                            alpha=0.5)


            kwargs = { 'alpha': 0.8 }

        ylim = autoscale.get_autoscale_y(axs[0], ylog=self.log_flux)
        for ax in axs:
            print "ylim: ", ylim
            yl = autoscale.get_autoscale_y(ax, ylog=self.log_flux, top_margin=0.2, bottom_margin=0.1)
            ylim = [min(ylim[0], yl[0]), max(ylim[1], yl[1])]

        for ax in axs:
            ax.set_ylim(ylim)

        if self.show_residual:
            data_flux_ = data_flux[data_wl==model_wl]
            data_mask_ = data_mask[data_wl==model_wl]
            data_flux_err_ = data_flux_err[data_wl==model_wl]
            residual = (data_flux_-median_flux)/data_flux_
            residual_err = (1./data_flux_ - (data_flux_-median_flux)/data_flux_**2) * data_flux_err_

            ymax = np.amax(abs(residual))
            ymax += 0.2*ymax
            ymax = 1.1

            colors = ["darkgreen"]
            indices = np.arange(len(data_wl))
            if has_mask:
                colors = ["darkgreen", mask_color]

            for ax in residual_axs:
                ax.set_ylim([-ymax, ymax])
                kwargs = {'alpha':0.7}

                ax.plot(ax.get_xlim(), [0.,0.],
                        color="black",
                        lw=0.7,
                        zorder=5,
                        **kwargs)

                for i, col in enumerate(colors):

                    if i==0:
                        ind = indices[data_mask_]
                    elif i==1:
                        ind = indices[~data_mask_]

                    ax.errorbar(model_wl[ind],
                            residual[ind],
                            yerr=residual_err[ind],
                            color = col,
                            ls=" ",
                            elinewidth = 0.5,
                            marker='o',
                            ms=3,
                            **kwargs
                            )

        for ax in axs:

            # Location of printed text
            x0, x1 = ax.get_xlim()
            x = x0 + (x1-x0)*0.03
            y0, y1 = ax.get_ylim()
            y = y1 - (y1-y0)*0.10

            # Label emission lines
            if self.plot_line_labels:
                x0, x1 = ax.get_xlim()
                n = 3
                i = 0
                prev_x = 0.
                y_shift = 0.050
                y_start = 0.075
                for key, label in self.line_labels.iteritems():

                    if self.wl_rest:
                        x = label["wl"]/wl_factor
                    else:
                        x = label["wl"]/wl_factor * (1.+self.observed_spectrum.data['redshift'])

                    if x < x0 or x > x1:
                        continue

                    if self.resolution is not None:
                        if abs(x-prev_x) < (x/self.resolution):
                            continue

                    if i >= n:
                        i=0

                    prev_x = x
                    y0, y1 = ax.get_ylim() 
                    if self.log_flux:
                        dy = np.log10(y1)-np.log10(y0)
                        y = 10.**(np.log10(y1)-dy*y_start-dy*y_shift*i)
                    else:
                        dy = y1-y0
                        y = y1 - dy*y_start-dy*y_shift*i

                    ax.text(x, y, 
                            label["label"], 
                            fontsize=14, 
                            rotation=45,
                            ha='center')

                    if self.log_flux:
                        y = 10.**(np.log10(y)-dy*y_shift)
                    else:
                        y -= dy*y_shift

                    ax.plot([x,x], [y,0.], 
                            ls="--",
                            lw=0.5,
                            color="black",
                            zorder=0)
                    i+=1


            if print_text:

                # Print the evidence
                try:
                    ax.text(x, y, "$\log(Z)=" + "{:.2f}".format(self.logEvidence) + "$", fontsize=10 )
                except AttributeError:
                    print "ciao"

                # Print the average reduced chi-square
                try:
                    aver_chi_square = self.PPC.data['aver_chi_square'][self.PPC.data['ID'] == ID]
                    y = y1 - (y1-y0)*0.15
                    ax.text(x, y, "$\langle\chi^2\\rangle=" + "{:.2f}".format(aver_chi_square) + "$", fontsize=10 )
                except AttributeError:
                    print "`PosteriorPredictiveChecks` not computed/loaded, hence " \
                    "<chi^2> for the object `" + str(ID) + "` is not available"

                try:
                    aver_red_chi_square = self.PPC.data['aver_red_chi_square'][self.PPC.data['ID'] == ID]
                    n_data = self.PPC.data['n_used_bands'][self.PPC.data['ID'] == ID]
                    y = y1 - (y1-y0)*0.20
                    ax.text(x, y,
                            "$\langle\chi^2/(\\textnormal{N}_\\textnormal{data}-1)\\rangle=" \
                            + "{:.2f}".format(aver_red_chi_square) + "\; \
                            (\\textnormal{N}_\\textnormal{data}=" + \
                            "{:d}".format(n_data) + ")" + "$", fontsize=10 )
                except AttributeError:
                    print "`PosteriorPredictiveChecks` not computed/loaded, hence " \
                    "<chi^2_red> for the object `" + str(ID) + "` is not available"

            if y0 < 0.: plt.plot( [x0,x1], [0.,0.], color='gray', lw=1.0 )

        #plt.tick_params(labelcolor='none', top='off', bottom='off', left='off', right='off')
        #plt.xlabel(
        #plt.ylabel("$F_{\\lambda} / (\\textnormal{erg} \, \
        #        \\textnormal{s}^{-1} \, \\textnormal{cm}^{-2} \, \
        #        \\textnormal{\AA}^{-1})$", va='center', rotation='vertical')

        #fig.text(0.5, 0.02, , ha='center')
        #fig.text(0.04, 0.5, 
        #        \\textnormal{s}^{-1} \, \\textnormal{cm}^{-2} \, \
        #        \\textnormal{\AA}^{-1})$", va='center', rotation='vertical')


        #plt.tight_layout()

        name = prepare_plot_saving(plot_name)

        fig.savefig(name, dpi=None, facecolor='w', edgecolor='w',
                orientation='portrait', papertype='a4', format="pdf",
                transparent=False, bbox_inches="tight", pad_inches=0.1)
        plt.close(fig)

        hdulist.close()



from beagle_utils import *
from beagle_parsers import standard_parser
from beagle_filters import PhotometricFilters
from beagle_photometry import Photometry
from beagle_pdf import PDF
from beagle_spectra import Spectrum
from beagle_summary_catalogue import BeagleSummaryCatalogue
from beagle_mock_catalogue import BeagleMockCatalogue
from beagle_residual_photometry import ResidualPhotometry
from beagle_multinest_catalogue import MultiNestCatalogue
from beagle_posterior_predictive_checks import PosteriorPredictiveChecks



#!/usr/bin/env python

from numpy import arange, array, bincount, ndarray, ones, where
from numpy.random import seed, random, randint

__author__ = "Tamas Nepusz, Denis Bzowy"
__version__ = "27jul2011"

class WalkerRandomSampling(object):
    """Walker's alias method for random objects with different probablities.
    
    Based on the implementation of Denis Bzowy at the following URL:
    http://code.activestate.com/recipes/576564-walkers-alias-method-for-random-objects-with-diffe/
    """
    
    def __init__(self, weights, keys=None, rand_seed=None):
        """Builds the Walker tables ``prob`` and ``inx`` for calls to `random()`.
        The weights (a list or tuple or iterable) can be in any order and they
        do not even have to sum to 1."""
        n = self.n = len(weights)
        if keys is None:
            self.keys = keys
        else:
            self.keys = array(keys)

        if rand_seed is not None:
            seed(rand_seed)

        if isinstance(weights, (list, tuple)):
            weights = array(weights, dtype=float)
        elif isinstance(weights, ndarray):
            if weights.dtype != float:
                weights = weights.astype(float)
        else:
            weights = array(list(weights), dtype=float)

        if weights.ndim != 1:
            raise ValueError("weights must be a vector")

        weights = weights * n / weights.sum()

        inx = -ones(n, dtype=int)
        short = where(weights < 1)[0].tolist()
        long = where(weights > 1)[0].tolist()
        while short and long:
            j = short.pop()
            k = long[-1]

            inx[j] = k
            weights[k] -= (1 - weights[j])
            if weights[k] < 1:
                short.append( k )
                long.pop()

        self.prob = weights
        self.inx = inx

    def random(self, count=None):
        """Returns a given number of random integers or keys, with probabilities
        being proportional to the weights supplied in the constructor.

        When `count` is ``None``, returns a single integer or key, otherwise
        returns a NumPy array with a length given in `count`.
        """
        if count is None:
            u = random()
            j = randint(self.n)
            k = j if u <= self.prob[j] else self.inx[j]
            return self.keys[k] if self.keys is not None else k

        u = random(count)
        j = randint(self.n, size=count)
        k = where(u <= self.prob[j], j, self.inx[j])
        return self.keys[k] if self.keys is not None else k


if __name__ == "__main__":
    # little examples, self-contained --
    N = 5
    Nrand = 1000
    randomseed = 1

    if randomseed:
        seed(randomseed)

    print Nrand, "Walker random sampling with weights .1 .2 .3 .4:"
    wrand = WalkerRandomSampling(arange(1, N))
    nrand = bincount(wrand.random(Nrand)).tolist()
    s = str(nrand)
    print s
    if N==5 and Nrand==1000 and randomseed==1:
        assert s == "[97, 207, 316, 380]"

    print Nrand, "Walker random sampling, strings with weights .1 .2 .3 .4:"
    abcd = dict(A=1, D=4, C=3, B=2)
    wrand = WalkerRandomSampling(abcd.values(), abcd.keys())
    nrand = defaultdict(int)
    for sample in wrand.random(Nrand):
        nrand[sample] += 1
    s = str(sorted(nrand.iteritems()))
    print s
    if N==5 and Nrand==1000 and randomseed==1:
        assert s == "[('A', 85), ('B', 199), ('C', 343), ('D', 373)]"



import os
import logging
from collections import OrderedDict
import json
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import rcParams
from scipy.interpolate import interp1d
from astropy.io import fits
from  itertools import izip_longest

import bokeh.plotting as bk_plt
import bokeh.models as bk_mdl

from beagle_utils import prepare_data_saving, prepare_plot_saving, \
        BeagleDirectories, is_FITS_file, data_exists, plot_exists, set_plot_ticks, \
        is_integer, match_ID

def grouper(n, iterable, fillvalue=None):
    "grouper(3, 'ABCDEFG', 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return izip_longest(fillvalue=fillvalue, *args)

def test_dim(testlist, dim=0):
   """tests if testlist is a list and how many dimensions it has
   returns -1 if it is no list at all, 0 if list is empty 
   and otherwise the dimensions of it"""
   if isinstance(testlist, list):
      if testlist == []:
          return dim
      dim = dim + 1
      dim = test_dim(testlist[0], dim)
      return dim
   else:
      if dim == 0:
          return -1
      else:
          return dim

def extract_err(err, data):
    '''private function to compute error bars
    Parameters
    ----------
    err : iterable
        xerr or yerr from errorbar
    data : iterable
        x or y from errorbar
    '''
    if (iterable(err) and len(err) == 2):
        a, b = err
        if iterable(a) and iterable(b):
            # using list comps rather than arrays to preserve units
            low = [thisx - thiserr for (thisx, thiserr)
                   in cbook.safezip(data, a)]
            high = [thisx + thiserr for (thisx, thiserr)
                    in cbook.safezip(data, b)]
            return low, high
    # Check if xerr is scalar or symmetric. Asymmetric is handled
    # above. This prevents Nx2 arrays from accidentally
    # being accepted, when the user meant the 2xN transpose.
    # special case for empty lists
    if len(err) > 1:
        fe = safe_first_element(err)
        if not ((len(err) == len(data) and not (iterable(fe) and
                                                len(fe) > 1))):
            raise ValueError("err must be a scalar, the same "
                             "dimensions as x, or 2xN.")
    # using list comps rather than arrays to preserve units
    low = [thisx - thiserr for (thisx, thiserr)
           in cbook.safezip(data, err)]
    high = [thisx + thiserr for (thisx, thiserr)
            in cbook.safezip(data, err)]
    return low, high

def errorbar(fig, x, y, xerr=None, yerr=None, kwargs={}):

  if xerr is not None:
      left, right = extract_err(xerr, x)
      fig.multi_line(left, right, **kwargs)

  if yerr is not None:
      lower, upper = extract_err(yerr, y)
      fig.multi_line(lower, upper, **kwargs)

class BeagleMockCatalogue(object):

    def __init__(self, params_file, 
            ID_key="ID", 
            file_name=None,
            ignore_string=None, 
            overwrite_plots=True,
            plot_title=None,
            n_bins=10):

        # Names of parameters, used to label the axes, whether they are log or
        # not, and possibly the extension name and column name containing the
        # "mock" parameters, i.e. the "true" parameters of the galaxy
        with open(params_file) as f:    
            # The use of the "OrderedDict" ensures that the order of the
            # entries in the dictionary reflects the order in the file
            self.adjust_params = json.load(f, object_pairs_hook=OrderedDict)

        self.ID_key = ID_key

        self.ignore_string = ignore_string

        self.overwrite_plots = overwrite_plots

        self.n_bins = n_bins

        self.plot_title = None
        if plot_title is not None:
            self.plot_title = plot_title
            if rcParams['text.usetex']:
                self.plot_title = self.plot_title.replace('_', '\_')

        if file_name is not None:
            self.load(file_name)

    def load(self, file_name=None):
        """ 
        Load a 'BEAGLE mock catalogue'

        Parameters
        ----------
        file_name : str
            Name of the file containing the catalogue.
        """

        if file_name is None:
            file_name = "BEAGLE_mock_catalogue.fits"

        self.file_name = os.path.basename(file_name)

        logging.info("Loading the `BeagleMockCatalogue` file: " + file_name)

        name = file_name
        if not os.path.dirname(file_name):
            name = os.path.join(BeagleDirectories.results_dir, 
                    BeagleDirectories.pypbeagle_data, 
                    file_name)

        self.hdulist = None

        if is_FITS_file(name):
            hdulist = fits.open(name)
            if len(hdulist) > 2:
                self.hdulist = hdulist
            else:
                self.data = fits.open(name)[1].data
                self.columns = fits.open(name)[1].columns
        else:
            self.data = ascii.read(name, Reader=ascii.basic.CommentedHeader)

    def get_param_values(self, ID, names):

        values = np.zeros(len(names), dtype=np.float32)

        for i, name in enumerate(names):

            if "extName" in self.adjust_params[name]:
                extName = self.adjust_params[name]["extName"]
            else:
                extName = "POSTERIOR PDF"

            if "colName" in self.adjust_params[name]:
                colName = self.adjust_params[name]["colName"]
            else:
                colName = name


            if self.hdulist is not None:
                data = self.hdulist[extName].data
            else:
                data = self.data

            if self.ID_key in data.dtype.names:
                if self.ignore_string is not None:
                    values[i] = data[colName][data[self.ID_key] == ID.replace(self.ignore_string, '')]
                else:
                    values[i] = data[colName][data[self.ID_key] == ID]
            else:
                if is_integer(ID):
                    row = int(ID)
                else:
                    row = int(ID.split('_')[0])

                values[i] = data[colName][row-1]

        return values

    def compute(self, file_list, file_name=None, overwrite=False):
        """ 
        """ 

        if file_name is None:
            file_name = "BEAGLE_mock_catalogue.fits"

        # Check if the `file_name` already exists
        if data_exists(file_name) and not overwrite:
            logging.warning('The file `' + file_name + '` already exists. \n Exiting the function.')
            return

        # We save in a dictionary the extension name and column name containing
        # the "true" parameters
        params_dict = dict()
        # This cycles over all keys containing the parameter names
        for key, value in self.adjust_params.iteritems():
            d = { "extName" : "POSTERIOR PDF", "colName" : key}
            log = {"log" : False}
            log.update(value)
            # This cycles over the dictioary items for one parameter
            for in_key, in_value in value.iteritems():
                if in_key == "mock":
                    # This will merge the default dictionary `d` with the
                    # one found in the json file
                    d.update(in_value)

            params_dict[key] = d

        # Read all FITS file in the `file_list`, and extract the columns
        # defined in the `params_dict` dictionary
        data = OrderedDict()
        n_files = len(file_list)
        data['ID'] = np.chararray(n_files, itemsize=20)
        for i, file in enumerate(file_list):
            hdulist = fits.open(os.path.join(BeagleDirectories.results_dir, file))
            data['ID'][i] = os.path.basename(file).split('_BEAGLE')[0]
            for key, value in params_dict.iteritems():
                val = hdulist[value["extName"]].data[value["colName"]]
                if not key in data:
                    data[key] = np.zeros(n_files)
                data[key][i] = val
            hdulist.close()

        # Initialize a new (empty) primary HDU for your output FITS file
        hdulist = fits.HDUList(fits.PrimaryHDU())
    
        new_columns = list()
        for key, value in data.iteritems():

            # The `ID` column contains a string, while all the other columns real data
            if 'ID' in key:
                new_columns.append(fits.Column(name=str(key), format='20A', array=data[key]))
            else:
                new_columns.append(fits.Column(name=str(key), format='E', array=data[key]))

        cols_ = fits.ColDefs(new_columns)
        new_hdu = fits.BinTableHDU.from_columns(cols_)

        # And finally append the newly created binary table to the hdulist
        # that will be printed to the ouput FITS file
        hdulist.append(new_hdu)

        name = prepare_data_saving(file_name, overwrite=overwrite)
        hdulist.writeto(name, clobber=overwrite)

        self.columns = new_hdu.columns
        self.data = new_hdu.data

    def compare_hist(self, summary_catalogue,
            class_indices=None,
            class_colors=None,
            plot_name=None,
            summary_type='median',
            percentile=95.,
            params_to_plot=None,
            pow=False,
            overwrite=False,
            kwargs=None):
            

        # Match IDs in the two catalogs
        indx1, indx2  = match_ID(self.hdulist[1].data['ID'], summary_catalogue.hdulist[1].data['ID'], ignore_string=self.ignore_string)

        #print "indx1: ", indx1, len(self.hdulist[1].data['ID']), max(indx1), len(indx1)
        #print "indx2: ", indx2, len(summary_catalogue.hdulist[1].data['ID']), max(indx2), len(indx2)

        # The summary statistics can be only 'mean' or 'median'
        if summary_type not in ('mean', 'median'):
            raise ValueError("`summary_type` con only be set to `mean` or `median`")

        if plot_name is None:
            plot_name = "BEAGLE_mock_retrieved_params_hist.pdf"

        # Check if the plot already exists
        if plot_exists(plot_name) and not overwrite:
            logging.warning('The plot `' + plot_name + '` already exists. \n Exiting the function.')
            return

        # By default you plot all parameters
        if params_to_plot is None:
            params_to_plot = list()
            for key, value in self.adjust_params.iteritems():
                params_to_plot.append(key)

        _n = int(np.ceil(np.sqrt(len(params_to_plot))))

        fig, axs = plt.subplots(_n, _n)
        fig.subplots_adjust(left=0.08, bottom=0.08, hspace=0.4)
        fontsize = 8

        axs = np.ravel(axs)

        diff_median = dict()
        diff_dispersion = dict()
        hist = list()
        edges = list()
        center = list()
        width = list()

        if kwargs is None:
            kwargs = {'alpha' : 0.7}

        for i, param in enumerate(params_to_plot):
            # Extract the row corresponding to `param` from the mock catalogue
            # (this array contains the "true") values of the parameters

            if "extName" in self.adjust_params[param]:
                extName = self.adjust_params[param]["extName"]
            else:
                extName = "POSTERIOR PDF"

            if "colName" in self.adjust_params[param]:
                colName = self.adjust_params[param]["colName"]
            else:
                colName = param

            true_param = self.hdulist[extName].data[colName][indx1]

            _col = colName + '_' + summary_type

            retrieved_param = summary_catalogue.hdulist[extName].data[_col][indx2]

            ax = axs[i]

            # Set the x- and y-axis labels
            label = self.adjust_params[param]['label']

            # Check if we need to compute the log10 of the param or not
            x = true_param
            y = retrieved_param

            if 'log' in self.adjust_params[param]:
                if self.adjust_params[param]['log'] and pow and '\log' in label:
                    label = label.replace('\log', '')
                if self.adjust_params[param]['log'] and not pow:
                    x = np.log10(x)
                    y = np.log10(y)

            xlabel = label + " (true-retrieved)"
            ylabel = '\# of objects'
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)

            diff = (x-y)

            if class_indices is None:
                class_indices = np.arange(len(diff))
                class_indices = [class_indices, ]

            diff_med = list()
            diff_disp = list()

            for i, indices in enumerate(class_indices):
                ind_set = set(indices)
                #print "indices: ", indices, len(diff)
                idx = [j for j, item in enumerate(indx1) if item in ind_set]
                indx = idx
                #print "indx: ", indx

                if 'diff_range' in self.adjust_params[param]:
                    range = self.adjust_params[param]['diff_range']
                else:
                    range = np.percentile(diff[idx], (0.5*(100.-percentile), 100.-0.5*(100.-percentile)))

                if class_colors is None:
                    color = None
                else:
                    color = class_colors[i]

                med, d0, d1 = np.percentile(diff[idx], [50., 0.5*(100.-68.), 100.-0.5*(100.-68.)])
                dispersion = 0.5*(d1-d0)

                diff_med.append(med)
                diff_disp.append(dispersion)

                ax.hist(diff[idx],
                    bins=self.n_bins,
                    range=range,
                    color=color,
                    lw=0,
                    **kwargs)

                ax.text(0.1+i*0.25, 1.05, "$\sigma=" + "{:.3f}".format(dispersion) + "$", 
                    fontsize=8, 
                    horizontalalignment='center', 
                    transform=ax.transAxes)

                ax.text(0.1+i*0.25, 1.14, "$\\textnormal{med}=" + "{:.3f}".format(med) + "$", 
                    fontsize=8, 
                    horizontalalignment='center', 
                    transform=ax.transAxes)

            diff_median[param] = diff_med
            diff_dispersion[param] = diff_disp

            for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] +
                         ax.get_xticklabels() + ax.get_yticklabels()):
                item.set_fontsize(fontsize)

            set_plot_ticks(ax, n_x=4, n_y=4)

        # Make the unused axes invisibles
        for ax in axs[len(params_to_plot):]:
            ax.axis('off')
        
        name = prepare_plot_saving(plot_name, overwrite=overwrite)

        #plt.tight_layout()

        fig.savefig(name, dpi=None, facecolor='w', edgecolor='w',
                orientation='portrait', papertype='a4', format="pdf",
                transparent=False, bbox_inches="tight", pad_inches=0.1)

        plt.close(fig)

        return diff_median, diff_dispersion

    def compare(self, summary_catalogue, 
            plot_name=None,
            summary_type='median',
            level=68.,
            params_to_plot=None,
            overwrite=False,
            rows=None, 
            interactive=False):
            

        # Match IDs in the two catalogs
        indx1, indx2  = match_ID(self.hdulist[1].data['ID'], summary_catalogue.hdulist[1].data['ID'], ignore_string=self.ignore_string)

        # Check whether the IDs of the two catalogues match
        #if 'ID' in self.hdulist[1].data:
        #    for i, ID in enumerate(self.hdulist[1].data['ID']):
        #        if not ID == summary_catalogue.hdulist['POSTERIOR PDF'].data['ID'][i]:
        #            raise Exception("The object IDs of the `mock` and `summary` catalogues do not match!")

        # The summary statistics can be only 'mean' or 'median'
        if summary_type not in ('mean', 'median'):
            raise ValueError("`summary_type` con only be set to `mean` or `median`")

        if plot_name is None:
            plot_name = "BEAGLE_mock_retrieved_params.pdf"

        # Check if the plot already exists
        if plot_exists(plot_name) and not overwrite:
            logging.warning('The plot `' + plot_name + '` already exists. \n Exiting the function.')
            return

        # By default you plot all parameters
        if params_to_plot is None:
            params_to_plot = list()
            for key, value in self.adjust_params.iteritems():
                params_to_plot.append(key)

        # Do you consider only some rows in the catalogue?
        #if rows is None:
        #    rows = np.arange(len(self.hdulist[1].data.field(0)))

        _n = int(np.ceil(np.sqrt(len(params_to_plot))))

        ######################################################################
        # If interactive is `True`, you draw an interactive plot instead of
        # plotting into a pdf file
        ######################################################################
        if interactive:

            # Size (in pixels) of each Bokeh figure
            size = 400

            # Name of the output html file created by Bokeh
            name = prepare_plot_saving(os.path.splitext(plot_name)[0]+'.html', 
                    overwrite=overwrite)

            # Tell Bokeh to save the plot into an output html file 
            bk_plt.output_file(name)

            # create a column data source for the plots to share, and fill the
            # dictionary with all the data that will then be used in the Bokeh
            # plot
            data = dict()
            data['ID'] = self.hdulist[1].data['ID'][indx1]
            for param in params_to_plot:

                # Store the "true" parameter
                key = param + '_true'
                data[key] = self.data[param][indx1]

                # and the retrieved one
                key = param + '_retrieved'
                _col = param + '_' + summary_type
                data[key] = summary_catalogue.hdulist['POSTERIOR PDF'].data[_col][indx2]

                # Get the errors
                _col = param + '_' + '{:.2f}'.format(level)
                tmp = summary_catalogue.hdulist['POSTERIOR PDF'].data[_col][indx2]

                # Store the errors in a way that is easily usable by the Bokeh
                # `multi_line` glyph
                err_xs = list()
                err_ys = list()
                key = param + '_true'
                for i, _x in enumerate(data[key]):
                    err_xs.append((_x, _x))
                    err_ys.append((tmp[i,0], tmp[i,1]))

                key = param + '_err_xs'
                data[key] = err_xs
                key = param + '_err_ys'
                data[key] = err_ys
                #data[key] = 0.5*(tmp[:,1]-tmp[:,0])

            strID = [ str(ID) + 
                    '_BEAGLE_input_for_Pierre.fits_snr_PS_CLEAR_PRISM_BEAGLE_marginal_SED_spec.pdf' 
                    for ID in data['ID']]

            data['strID'] = strID

            # Build the `ColumnDataSource` (a sort of dictionary) that will be
            # used by Bokeh to extract the data
            source = bk_mdl.ColumnDataSource(data=data)

            # Define the "tools" that will be included in the interactive Bokeh plot
            tools = 'wheel_zoom,pan,reset,resize,hover,tap'

            figs = list()
            for param in params_to_plot:
                # create a new plot and add a renderer
                fig = bk_plt.figure(tools=tools, width=size, height=size, title=None)

                # Plot the x and y data points as circles
                x = param + '_true'
                y = param + '_retrieved'
                error_low = param + '_err_xs'
                error_up = param + '_err_ys'
                fig.circle(x, y, source=source)

                # Plot the errorbars
                fig.multi_line(xs=error_low, ys=error_up, source=source, alpha=0.6)

                # Plot the 1-to-1 relation
                _max = np.amax(np.ravel([data[x], data[y]]))
                _min = np.amin(np.ravel([data[x], data[y]]))
                fig.line([_min, _max], [_min, _max], color='red')

                # Label the x- and y-axis
                label = self.adjust_params[param]['label_plain']

                xlabel = label + " (true)"
                fig.xaxis.axis_label = xlabel

                ylabel = label
                fig.yaxis.axis_label = ylabel
                    
                # Append the newly created figure to the `fgs` list of Bokeh figures
                figs.append(fig)

            # Arrange the different figures in a matrix-list
            grid_figs = list()

            for i in range(0,_n):
                _tmp = list()
                for j in range(_n):
                    if i*_n+j >= len(figs):
                        break
                    else:
                        _tmp.append(figs[i*_n+j])
                grid_figs.append(_tmp)

            # Use the matrix-list to create a grid of Bokeh figures
            p = bk_plt.gridplot(grid_figs)

            # Here we customize the behaviour of some tools
            hover = p.select(dict(type=bk_mdl.HoverTool))
            
            hover.tooltips = [
                 ("(x,y)", "($x, $y)"),
                ('ID', '@ID'),
            ]

            path = os.path.join(os.getcwd(),
                    '@strID')
            url = "file://" + path
            print "url: ", url
            taptool = p.select(type=bk_mdl.TapTool)
            taptool.callback = bk_mdl.OpenURL(url=url)

            bk_plt.show(p)

            return

        ######################################################################
        # Below is a standard plot written to a pdf file
        ######################################################################

        fig, axs = plt.subplots(_n, _n)
        fig.subplots_adjust(left=0.08, bottom=0.08)
        fontsize = 8

        axs = np.ravel(axs)

        for i, param in enumerate(params_to_plot):
            # Extract the row corresponding to `param` from the mock catalogue
            # (this array contains the "true") values of the parameters

            if "extName" in self.adjust_params[param]:
                extName = self.adjust_params[param]["extName"]
            else:
                extName = "POSTERIOR PDF"

            if "colName" in self.adjust_params[param]:
                colName = self.adjust_params[param]["colName"]
            else:
                colName = param

            true_param = self.hdulist[extName].data[colName][indx1]

            _col = colName + '_' + summary_type

            retrieved_param = summary_catalogue.hdulist[extName].data[_col][indx2]

            _col = colName + '_' + '{:.2f}'.format(level)
            error = summary_catalogue.hdulist[extName].data[_col][indx2]

            ax = axs[i]

            # Set the x- and y-axis labels
            label = self.adjust_params[param]['label']
            xlabel = label + " (true)"
            ylabel = label

            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)

            # Check if we need to compute the log10 of the param or not
            x = true_param
            y = retrieved_param
            yerr_u = error[:,1]-y
            yerr_d = y-error[:,0]

            is_log = False
            if 'log' in self.adjust_params[param]:
                if self.adjust_params[param]['log']:
                    is_log = True
                    #x = np.log10(x)
                    #y = np.log10(y)
                    #yerr_d = np.log10(yerr_d)
                    #yerr_u = np.log10(yerr_u)

            ax.errorbar(x, 
                    y, 
                    yerr=[yerr_d, yerr_u],
                    ls="",
                    marker='o',
                    ms=3, 
                    mew=0,
                    elinewidth=0.6,
                    capsize=3,
                    alpha=0.5
                    )

            for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] +
                         ax.get_xticklabels() + ax.get_yticklabels()):
                item.set_fontsize(fontsize)

            # Make the axes in log-scale
            if is_log:
                ax.set_xscale('log')
                ax.set_yscale('log')
            else:
                set_plot_ticks(ax, n_x=4, n_y=4)

            # Make axes of same aspect ratio
            plt.axis('equal')

            # Set same axis range for x- and y-axis, and use the (optional)
            # user-provided value  
            xx = ax.get_xlim()
            yy = ax.get_ylim()
            ll = [min([xx[0], yy[0]]), max(xx[1], yy[1])]
            if 'axis_range' in self.adjust_params[param]:
                rr = self.adjust_params[param]['axis_range']
                ll = [max(ll[0],rr[0]), min(ll[1],rr[1])]
                ax.set_autoscale_on(False)

            ax.set_xlim(ll)
            ax.set_ylim(ll)

            # Plot the 1-to-1 relations
            ax.plot(ax.get_xlim(), ax.get_ylim(), 
                    ls="-",
                    color="black",
                    c=".3")

        # Make the unused axes invisibles
        for ax in axs[len(params_to_plot):]:
            ax.axis('off')
        
        name = prepare_plot_saving(plot_name, overwrite=overwrite)

        plt.tight_layout()

        fig.savefig(name, dpi=None, facecolor='w', edgecolor='w',
                orientation='portrait', papertype='a4', format="pdf",
                transparent=False, bbox_inches="tight", pad_inches=0.1)

        plt.close(fig)


    def plot_input_param_distribution(self, savefig=True):

        n_par = len(self.adjust_params)
        fig, axs = plt.subplots(n_par, n_par)
        fig.subplots_adjust(left=0.08, bottom=0.08, hspace=0.2)
        fontsize = 8
        axes_linewidth = 0.7
        color = "grey"

        last = n_par-1

        for i, (keyX, valueX) in enumerate(self.adjust_params.iteritems()):

            if "extName" in self.adjust_params[keyX]:
                extNameX = self.adjust_params[keyX]["extName"]
            else:
                extNameX = "POSTERIOR PDF"

            if "colName" in self.adjust_params[keyX]:
                colNameX = self.adjust_params[keyX]["colName"]
            else:
                colNameX = keyX

            valuesX = self.hdulist[extNameX].data[colNameX]
            if 'log' in self.adjust_params[keyX]:
                if self.adjust_params[keyX]['log']:
                    valuesX = np.log10(valuesX)

            xlabel = self.adjust_params[keyX]['label']

            for j, (keyY, valueY) in enumerate(self.adjust_params.iteritems()):

                if "extName" in self.adjust_params[keyY]:
                    extNameY = self.adjust_params[keyY]["extName"]
                else:
                    extNameY = "POSTERIOR PDF"

                if "colName" in self.adjust_params[keyY]:
                    colNameY = self.adjust_params[keyY]["colName"]
                else:
                    colNameY = keyY

                valuesY = self.hdulist[extNameY].data[colNameY]
                if 'log' in self.adjust_params[keyY]:
                    if self.adjust_params[keyY]['log']:
                        valuesY = np.log10(valuesY)

                ylabel = self.adjust_params[keyY]['label']

                ax = axs[i][j]

                if j > i:
                    ax.axis('off')
                    continue

                elif i == j:

                    ax.hist(valuesX,
                        bins=self.n_bins,
                        lw=0,
                        color=color)

                elif j < i:

                    ax.plot(valuesY,
                            valuesX,
                            marker='o',
                            ms=2, 
                            color=color,
                            ls="")


                    #ax.set_xlim(rangeY)
                    #ax.set_ylim(rangeX)

                if i == j:
                    ax.yaxis.tick_right()
                    if i == 0:
                        ax.set_xticklabels(('',))
                        ax.set_ylabel(xlabel)
                    elif i == last:
                        ax.set_xlabel(ylabel)
                    else:
                        ax.set_xticklabels(('',))
                elif j == 0:
                    ax.set_ylabel(xlabel)
                    if i != last:
                        ax.set_xticklabels(('',))
                elif i == last:
                    ax.set_xlabel(ylabel)
                    ax.set_yticklabels(('',))
                else:
                    ax.set_xticklabels(('',))
                    ax.set_yticklabels(('',))


                #if i == last:
                #    ax.set_xlabel(ylabel)
                #    if j != 0 and j != last:
                #        ax.set_yticklabels(('',))
                #    elif j == last:
                #        ax.yaxis.tick_right()

                #ax.set_ylabel('j ' + str(j) + 'i ' + str(i) )
                #if i == 0:

                #if j != 0 and i != (n_par-1) and i != j:
                #    ax.set_xticklabels(('',))
                #    ax.set_yticklabels(('',))

                #if i == j and i != 0 and i != (n_par-1):
                #    if j != 0:
                #        ax.set_xticklabels(('',))

                #if j == 0:
                #    ax.set_ylabel(xlabel)
                #    if i != 0:
                #            ax.set_xticklabels(('',))


                for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] +
                             ax.get_xticklabels() + ax.get_yticklabels()):
                    item.set_fontsize(fontsize)

                set_plot_ticks(ax, n_x=3, n_y=3)

                ax.tick_params(which='minor', axis='both',
                                length=1.5, width=axes_linewidth)

                ax.tick_params(which='major', axis='both',
                                length=3, width=axes_linewidth)

                for axis in ['top','bottom','left','right']:
                  ax.spines[axis].set_linewidth(axes_linewidth)


        # Make the unused axes invisibles
        #for ax in axs[n_par:]:
        #    ax.axis('off')

        # Fine-tune figure; hide x ticks for top plots and y ticks for right plots
        plt.setp([a.get_xticklabels() for a in axs[0, :]], visible=False)
        plt.setp([a.get_yticklabels() for a in axs[:, 1]], visible=False)
    
        if self.plot_title is not None:
            fig.suptitle(self.plot_title)

        if savefig:

            plot_name = "BEAGLE_mock_input_params_distribution.pdf"
            name = prepare_plot_saving(plot_name, overwrite=self.overwrite_plots)

            #plt.tight_layout()

            fig.savefig(name, dpi=None, facecolor='w', edgecolor='w',
                    orientation='portrait', papertype='a4', format="pdf",
                    transparent=False, bbox_inches="tight", pad_inches=0.1)

        plt.close(fig)

        return fig, axs



import numpy as np
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt
import cPickle

import os
import sys
import dependencies.WeightedKDE as WeightedKDE

from beagle_utils import weighted_avg_and_std, BeagleDirectories, match_ID, prepare_plot_saving

class ResidualPhotometry(object):

    def load(self, file_name):
        """ 
        Load the pre-computed residul photometry from a cPickle file.

        Parameters
        ----------
        file_name : str
            Name of the file containing the cPicke dump.
        """ 

        name = os.path.join(BeagleDirectories.results_dir,
                BeagleDirectories.pypbeagle_data, file_name)

        file = open(name, 'rb')
        self.residual_kde_pdf = cPickle.load(file)
        file.close()

    def compute(self, observed_catalogue, beagle_summary_catalogue,
            filters, summary_stat=None, cPickleName=None):

        """ 
        Compute the residual photometry between an observed and model catalogues.  

        Parameters
        ----------
        observed_catalogue : `ObservedCatalogue` class object
            

        beagle_summary_catalogue : `BeagleSummaryCatalogue` class object
            

        filters : `PhotometricFilters` class object
            
        summary_stat : str, optional
            Either 'mean' or 'median', determines which type of summary
            statistics is used to compute the residual. By default the median
            of the marginal PDF is used.

        """

        if summary_stat is None:
            summary_stat = "median"

        beagle_data = beagle_summary_catalogue.hdulist['MARGINAL PHOTOMETRY'].data

        catalogue_data = observed_catalogue.data

        indx_beagle, indx_catalogue = match_ID(beagle_data['ID'], catalogue_data['ID'])

        beagle_data = beagle_data[indx_beagle]
        catalogue_data = catalogue_data[indx_catalogue]

        print "ID: ", beagle_data['ID']
        print "ID: ", catalogue_data['ID']

        # As a sanity check, check if the ID match among the two catalogues, by
        # random picking some indices here and there...
        #if (beagle_data['ID'][0] != catalogue_data['ID'][0]) or \
        #(beagle_data['ID'][-1] != catalogue_data['ID'][-1]):
        #    raise ValueError("The object IDs between the BEAGLE summary catalogue and \
        #        the observed catalogue do not match!")

        self.residual_kde_pdf = list() 

        jy = 1.E-26

        for i in range(filters.n_bands):

            obs_flux = catalogue_data[filters.data['flux_colName'][i]] * filters.units / jy
            obs_flux_err = catalogue_data[filters.data['flux_errcolName'][i]] * filters.units  / jy

            name = '_' + filters.data['label'][i] + '_'
            model_flux = beagle_data[name+'_'+summary_stat] / jy
            model_flux_err = 0.5 * (beagle_data[name+'_68.00'][:,1]-beagle_data[name+'_68.00'][:,0]) / jy

            mask = np.zeros(len(obs_flux), dtype=bool)

            mask[(obs_flux > 0.) & (obs_flux_err > 0.) & (model_flux > 0.) & (model_flux_err > 0.)] = True

            # Compute the residual in magnitude scale
            residual = -2.5*np.log10(model_flux[mask]/obs_flux[mask])

            # The residual error follows from standard error propagation
            residualErr = np.sqrt((2.5/model_flux[mask]*model_flux_err[mask])**2 + (2.5/obs_flux[mask]*obs_flux_err[mask])**2)

            # Compute the distribution of the residual by means of a weighted kernel density estimation
            # Sometimes at high redshift, short wavelength filters all contain model_flux = 0 when wavelength
            # of filter outside of wavelength range of templates, hence testing the length of residual
            if len(residual) > 0:
                kde_pdf = WeightedKDE.gaussian_kde(residual, weights = 1./residualErr)
            else:
                kde_pdf = 0

            self.residual_kde_pdf.append(kde_pdf)

        if cPickleName is not None:
            cPickle.dump(self.residual_kde_pdf, cPickleName, cPickle.HIGHEST_PROTOCOL)

    def plot(self, 
            filters, 
            plot_name="BEAGLE_residual_photometry.pdf", 
            x_range=(-1.,1.), 
            n_x=1000, 
            summary_stat="median"):
        """ 
        Plot the residual photometry.

        Parameters
        ----------
        filters : PhotometricFilters class object

        plot_name: str, optional
            File name of the output plot.

        x_range : iterable float size=2, optional
            Range over which computing the actual residual density function
            from the computed KDE residual.

        n_x : int, optional 
            Number of values over which computing the residual density
            function.

        summary_stat : str, optional
            Either 'mean' or 'median', determines which type of summary
            statistics is used.
            
        Notes
        -----
        The residuals density function is computed at each residual value
        defined by np.linspace(x_range[0], x_range[1], n_x)
        """

        # The grid of residual values over which computing the residual density
        # function
        x_grid = np.linspace(x_range[0], x_range[1], n_x)

        for residual_kde_pdf in self.residual_kde_pdf:

            # The residual density function, computed from the residual KDE
            kde_pdf_grid = np.array(residual_kde_pdf(x_grid))

            # You also compute the median, or mean residual, and its dispersion (or
            # 68 % credible region)
            cumul_pdf = np.cumsum(kde_pdf_grid)

            # Be sure the it sums up to 1
            cumul_pdf /= cumul_pdf[-1]

            # Compute interpolant of cumulative PDF (linear interpolation)
            interp_cumul_pdf = interp1d(cumul_pdf, x_grid)

            if "median" in summary_stat:
                med = interp_cumul_pdf(0.5)
                # This corresponds to a 68 % credible region
                interval = interp_cumul_pdf(0.84) - interp_cumul_pdf(0.16)
                print "\n median = {:.3f}".format(med)
                print "68 % interval = {:.3f}".format(interval)
            elif "mean" in summary_stat:
                mean, stddev = weighted_avg_and_std(x_grid, kde_pdf_grid)
                print "\n mean = {:.3f}".format(mean)
                print "stddev = {:.3f}".format(stddev)

        # Compute interpolant of cumulative PDF (linear interpolation)
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)

        ax.plot(x_grid, kde_pdf_grid , ls="-", color = "black")

        name = prepare_plot_saving(plot_name)

        fig.savefig(name, dpi=None, facecolor='w', edgecolor='w',
            orientation='portrait', papertype='a4', format="pdf",
            transparent=False, bbox_inches="tight", pad_inches=0.1)

        plt.close(fig)





import os
import logging
from collections import OrderedDict
import json
import numpy as np
from scipy.interpolate import interp1d
from astropy.io import fits

from beagle_utils import prepare_data_saving, BeagleDirectories, getPathForData, data_exists
from significant_digits import to_precision, float_nsf

def get1DInterval(param_values, probability, levels):

    """ 
    Compute several quantities from a 1D probability density function

    Parameters
    ----------
    param_values : numpy array
        Contains the values of the parameter.

    probability : numpy array 
        Contains the probability associated with each value of the
        parameter, hence must have same dimension as 'param_values'.

    levels : numpy array or list containing float
        Contains the (percentage) levels used to compute the credible
        regions, e.g. levels=[68.,95.] will compute 68 % and 95 % (central)
        credible regions

    Returns
    -------
    mean : float
        Mean of the parameter, computed as
        sum(probability*param_values)/sum(probability)
    median : float
        Median of the parameter, computed from the cumulative integral of
        the PDF
    interval : list of float
        2-dimensional list containing the lower and upper value of the
        parameter corresponding to the different `levels`

    """

    sort_ = np.argsort(param_values)

    # ******************************************************************
    # Here you must simply use `cumsum`, and not `cumtrapz` as in
    # beagle_utils.prepare_violin_plot, since the output of MultiNest are a set
    # of weights (which sum up to 1) associated to each set of parameters (the
    # `p_j` of equation 9 of Feroz+2009), and not a probability density (as the
    # MultiNest README would suggest).
    # ******************************************************************
    cumul_pdf = np.cumsum(probability[sort_])
    cumul_pdf /= cumul_pdf[len(cumul_pdf)-1]

    # Get the interpolant of the cumulative probability
    f_interp = interp1d(cumul_pdf, param_values[sort_])

    # You shoud integrate rather than summing here
    mean = np.sum(probability * param_values) / np.sum(probability)

    median = f_interp(0.5)

    interval = list()
    for lev in levels:

        low, high = f_interp([0.5*(1.-lev/100.), 1.-0.5*(1.-lev/100.)])
        interval.append([low,high])

    return mean, median, interval


class BeagleSummaryCatalogue(object):

    def __init__(self, file_name=None, config_file=None):

        if file_name is not None:
            self.file_name  = file_name
        else:
            self.file_name = "BEAGLE_summary_catalogue.fits" 

        if config_file is not None:
            if os.path.dirname(config_file) is None:
                self.config_file  = os.path.join(BeagleDirectories.results_dir, 
                        config_file)
            else:
                self.config_file = config_file
        else:
            self.config_file = os.path.join(BeagleDirectories.results_dir, 
                    "summary_config.json")

    def exists(self):

        return data_exists(self.file_name)

    def load(self):
        """ 
        Load a 'BEAGLE summary catalogue'

        Parameters
        ----------
        file_name : str
            Name of the file containing the catalogue.
        """

        logging.info("Loading the `BeagleSummaryCatalogue` file: " + self.file_name)

        name = getPathForData(self.file_name)
        self.hdulist = fits.open(name)

    def compute(self, file_list, overwrite=False, levels=[68.,95.]):
        """ 
        """ 

        with open(self.config_file) as f:    
            hdu_col = json.load(f, object_pairs_hook=OrderedDict)

        # You consider the first file in the list and use as a "mold" to create
        # the structure (binary tables and their columns) of the output FITS file
        firstfile = os.path.join(BeagleDirectories.results_dir, file_list[0])
        hdulist = fits.open(firstfile)

        n_objects = len(file_list)

        # Initialize a new (empty) primary HDU for your output FITS file
        self.hdulist = fits.HDUList(fits.PrimaryHDU())
    
        # Now you cycle over all extension and columns that you want to put in
        # the summary catalogue
        for hdu in hdu_col:
            new_columns = list()

            # Pick the extension name
            hdu_name = hdu['name']

            # The first column of each output extension contains the object ID
            #new_columns.append(fits.Column(name='ID', format='K'))
            new_columns.append(fits.Column(name='ID', format='20A'))

            # You just consider the columns defined in the structure
            if 'columns' in hdu:
                columnNames = hdu['columns']
            # While by default you take all columns in that extensions
            else:
                columnNames = hdulist[hdu_name].columns.names

            # For each column, you add a '_mean', '_median' and confidence
            # intervals columns, taking the appropriate units from the FITS
            # file that you are using as a mold
            for col_index in range(len(hdulist[hdu_name].columns)):

                col_ = hdulist[hdu_name].columns[col_index]

                if col_.name not in columnNames:
                    continue
    
                new_columns.append(fits.Column(name=col_.name+'_mean',
                    format=col_.format, unit=col_.unit))

                new_columns.append(fits.Column(name=col_.name+'_median',
                    format=col_.format, unit=col_.unit))

                for lev in np.array(levels):
                    new_columns.append(fits.Column(name=col_.name + '_' +
                        "{:.2f}".format(lev), format='2'+col_.format[-1],
                        unit=col_.unit))

            # Create the "column definition"
            cols_ = fits.ColDefs(new_columns)

            # Create the actual binary table, with the correct number of rows
            # to accomodate all objects
            nrows = len(file_list)
            new_hdu = fits.BinTableHDU.from_columns(cols_, nrows=nrows)
            new_hdu.name = hdu_name

            # And finally append the newly created bunary table to the hdulist
            # that will be printed to the ouput FITS file
            self.hdulist.append(new_hdu)

        hdulist.close()

        # Now you can go through each file, and compute the required quantities

        for i, file in enumerate(file_list):
            hdulist = fits.open(os.path.join(BeagleDirectories.results_dir, file))
            end = file.find('_' + BeagleDirectories.suffix)

            # Extract the object ID from the file_name
            #ID = np.int(np.float(os.path.basename(file[0:end])))
            ID = os.path.basename(file[0:end])

            probability = hdulist['posterior pdf'].data['probability']

            for hdu in hdu_col:
                hdu_name = hdu['name']
                if 'columns' in hdu:
                    columnNames = hdu['columns']
                else:
                    columnNames = hdulist[hdu_name].columns.names

                for col_name in columnNames:
                    self.hdulist[hdu_name].data['ID'][i] = ID
                    par_values = hdulist[hdu_name].data[col_name]

                    mean, median, interval = get1DInterval(par_values, probability, levels)

                    self.hdulist[hdu_name].data[col_name+'_mean'][i] = mean
                    self.hdulist[hdu_name].data[col_name+'_median'][i] = median

                    for j, lev in enumerate(levels):
                        levName = col_name + '_' + "{:.2f}".format(lev)
                        self.hdulist[hdu_name].data[levName][i] = interval[j]

            hdulist.close()

        name = prepare_data_saving(self.file_name)
        self.hdulist.writeto(name, clobber=overwrite)

    def extract_MAP_solution(self, file_list, overwrite=False):
        """ 
        """ 

        # Now you can go through each file, and extract the MAP solution
        for i, file in enumerate(file_list):

            # Open the original BEAGLE FITS file
            hdulist = fits.open(os.path.join(BeagleDirectories.results_dir, file))

            # Get the posterior probability
            post = hdulist['posterior pdf'].data['probability']

            # Maximum of the posterior PDF, i.e. you select the template
            # corresponding to the mode of the posterior distributions
            MAP_indx = np.argmax(post)

            # Now write in a separate FITS file the pararmeters corresponding to the MAP row
            new_hdulist = fits.HDUList(fits.PrimaryHDU())

            for hdu in hdulist:

                if hdu.data is None:
                    continue

                if hdu.is_image:
                    new_hdu = fits.PrimaryHDU()
                    new_hdu.name = hdu.name
                    new_hdu.data = hdu.data[MAP_indx,:]
                else:
                    new_hdu = fits.BinTableHDU.from_columns(hdu.columns, nrows=1)
                    new_hdu.name = hdu.name
                    if 'sed wl' in hdu.name.lower() or 'sed mask' in hdu.name.lower():
                        new_hdu.data = hdu.data
                    else:
                        new_hdu.data[0] = hdu.data[MAP_indx]

                new_hdulist.append(new_hdu)

            # Extract the object ID from the file_name
            end = file.find('_' + BeagleDirectories.suffix)
            ID = os.path.basename(file[0:end])

            file_name = ID + '_BEAGLE_MAP.fits.gz'
            name = prepare_data_saving(file_name)
            new_hdulist.writeto(name, clobber=overwrite)

            new_hdulist.close()
            hdulist.close()

    def make_latex_table(self, param_names, 
            IDs=None,
            level=68., 
            summary_statistics='median',
            significant_digits=2):

        n = significant_digits
        
        n_rows = len(self.hdulist[1].data['ID'])

        if IDs is None:
            IDs = self.hdulist[1].data['ID']

        for ID in IDs:
            print "\nID: ", ID, 
            row = np.arange(n_rows)[self.hdulist[1].data['ID'] == ID][0]
            for param in param_names:
                colName = param + '_' + summary_statistics
                for hdu in self.hdulist:
                    if hasattr(hdu, 'columns'):
                        if colName in hdu.columns.names:
                            sum_stat = hdu.data[colName][row]
                            errColName = param + '_' + "{:.2f}".format(level)
                            error = hdu.data[errColName][row]

                            print to_precision(sum_stat, n+2), '_{-' + to_precision(sum_stat-error[0], n) \
                                    + '}^{+' + to_precision(error[1]-sum_stat, n) + '} &',
                            break

                        



#!/usr/bin/env python
import sys
import os
import argparse
import re
import numpy as np
import ConfigParser
import logging
from matplotlib import rc
from astropy.io import ascii
from astropy.io import fits

from pathos.multiprocessing import ProcessingPool 
from pyp_beagle import *
from _version import __version__


def main():

    # Load the default command line argument parser
    parser = standard_parser()

    # Add package version
    parser.add_argument('-v', '--version', 
            action='version', 
            version='%(prog)s ' + __version__
            )

    # Get parsed arguments
    args = parser.parse_args()    

    # Set logging level
    logging.basicConfig(level=args.loglevel)

    # Set directory containing BEAGLE results files
    BeagleDirectories.results_dir = args.results_dir

    # Read parameter file
    config = ConfigParser.SafeConfigParser()

    # Check if you passed a name of the parameter file, otherwise search for a
    # suitable parameter file in the BEAGLE-input-files folder
    param_file = None
    if args.param_file is not None:
        param_file = args.param_file
    else:
        for file in os.listdir(os.path.join(BeagleDirectories.results_dir, 
            BeagleDirectories.beagle_input_files)):
            if file.endswith('.param'):
                file_name = os.path.join(BeagleDirectories.results_dir, 
                        BeagleDirectories.beagle_input_files, file)
                if '[main]' in open(file_name).readline():
                    param_file = file
                    break

    if param_file is None:
        raise ValueError("No parameter file has been specified, nor a parameter file can be " 
                "found in the Beagle results directory")

    # Set name of parameter file
    BeagleDirectories.param_file = param_file 

    # Parse the parameter file
    name = os.path.join(BeagleDirectories.results_dir, BeagleDirectories.beagle_input_files, 
            param_file)
    config.read(name)

    # Check if the parameter file contains a LIST OF SPECTRA
    has_spectra = config.has_option('main', 'LIST OF SPECTRA')

    # Check if the parameter file contains a PHOTOMETRIC CATALOGUE
    has_photometry = config.has_option('main', 'PHOTOMETRIC CATALOGUE')

    # Global font size
    font = {'size': 16}
    rc('font', **font)

    # Get list of results files and object IDs from the results directory
    file_list, IDs = get_files_list(suffix=args.suffix)
    if len(file_list) == 0:
        raise ValueError("No Beagle results files are present in the directory " + BeagleDirectories.results_dir)

    regex = None
    if args.regex_ignore is not None:
        #regex = re.compile(r"_MC\w+", re.IGNORECASE)
        regex = re.compile(args.regex_ignore, re.IGNORECASE)

    if args.ID_list is not None:
        IDs_ = IDs
        file_list_ = file_list
        for ID, file in zip(IDs_, file_list_):
            if regex is not None:
                ID_ = regex.sub('', ID)
            else:
                ID_ = ID
            if not ID_ in args.ID_list:
                IDs.remove(ID)
                file_list.remove(file)

    # Load mock catalogue
    mock_catalogue = None
    if args.mock_file_name is not None:

        # JSON file containing the configuration for the mock catalogue plots
        params_file = os.path.join(BeagleDirectories.results_dir, args.json_file_mock)
        mock_catalogue = BeagleMockCatalogue(params_file, ignore_string=regex, plot_title=args.plot_title)
        mock_catalogue.load(args.mock_file_name)

    # JSON file containing the parameters to be plotted in the triangle plot
    params_file = os.path.join(BeagleDirectories.results_dir, args.json_file_triangle)

    # Compute the summary catalogue
    summary_catalogue = BeagleSummaryCatalogue()
    if args.compute_summary:
        if not summary_catalogue.exists():
            summary_catalogue.compute(file_list)

    if args.extract_MAP:
        summary_catalogue.extract_MAP_solution(file_list)

    # Comparison plots of true vs retrieved values 
    if args.mock_file_name is not None:
        if not summary_catalogue.exists():
            summary_catalogue.compute(file_list)
        summary_catalogue.load()

        #mock_catalogue.plot_input_param_distribution()
        mock_catalogue.compare_hist(summary_catalogue, overwrite=True)
        mock_catalogue.compare(summary_catalogue, overwrite=True)

    # ---------------------------------------------------------
    # --------- Post-processing of photometric data -----------
    # ---------------------------------------------------------
    if has_photometry:

        # Initialize an instance of the main "Photometry" class
        my_photometry = Photometry(key=args.ID_key, 
                x_log=args.plot_log_wl, 
                plot_single_solution=args.plot_single_solution,
                plot_full_SED=args.plot_full_SED)

        # We can load a set of photometric filters
        try:
            filters_file = os.path.expandvars(config.get('main', 'FILTERS FILE'))
            filters_throughputs = None
        except:
            filters_file = os.path.expandvars(config.get('main', 'FILTERS CONFIGURATION'))
            filters_throughputs = os.path.expandvars(config.get('main', 'FILTERS THROUGHPUTS'))

        my_photometry.filters.load(filters_file, 
                filters_folder=args.filters_folder, 
                filters_throughputs=filters_throughputs)

        # Load observed photometric catalogue
        file_name = os.path.expandvars(config.get('main', 'PHOTOMETRIC CATALOGUE'))
        my_photometry.observed_catalogue.load(file_name)

    # ---------------------------------------------------------
    # -------- Post-processing of spectroscopic data ----------
    # ---------------------------------------------------------
    if has_spectra:

        # Initialize an instance of the main "Spectrum" class
        my_spectrum = Spectrum(params_file, 
                resolution=args.resolution, 
                plot_full_SED=args.plot_full_SED,
                wl_range=args.wl_range,
                line_labels_json=args.line_labels_json,
                plot_line_labels=args.plot_line_labels, 
                mock_catalogue=mock_catalogue,
                log_flux=args.plot_log_flux,
                show_residual=args.show_residual,
                print_ID=args.print_ID,
                wl_rest=args.wl_rest,
                draw_steps=args.draw_steps,
                plot_suffix=args.plot_suffix,
                wl_units=args.wl_units)

        my_spectrum.observed_spectrum.configure(config=config)


        # File containing list of input spectra
        inputSpectraFileName = os.path.expandvars(config.get('main', 'LIST OF SPECTRA'))

        lines = list()
        with open(inputSpectraFileName, 'r') as f:
            for line in f:
                # Get rid of the "\n" char at the end of the line
                line = line.strip()
                line = os.path.join(os.path.dirname(inputSpectraFileName), line)
                lines.append(line)

        file_names = list()

        for ID in IDs:
            ID_ = ID
            if regex is not None:
                ID_ = regex.sub('', ID_)
            for line in lines:
                line_ = trimFitsSuffix(os.path.basename(line))
                if regex is not None:
                    line_ = regex.sub('', line_)
                if ID_ == line_:
                    file_names.append(line)
                    break

    # Create "pool" of processes
    if args.np > 1:
        pool = ProcessingPool(nodes=args.np)

    # Plot the marginal SED
    if args.plot_marginal:
        if args.np > 1:
            if has_spectra:
                pool.map(my_spectrum.plot_marginal, IDs, file_names)

            if has_photometry:
                pool.map(my_photometry.plot_marginal, IDs)
        else:
            for i, ID in enumerate(IDs):
                if has_spectra:
                    my_spectrum.plot_marginal(ID, file_names[i])

                if has_photometry:
                    my_photometry.plot_marginal(ID)

    # Plot the triangle plot
    if args.plot_triangle:

        # Set parameter names and labels
        my_PDF = PDF(params_file, 
                mock_catalogue=mock_catalogue,
                plot_single_solution=args.plot_single_solution)

        if args.np > 1:
            pool.map(my_PDF.plot_triangle, IDs)
        else:
            for ID in IDs:
                my_PDF.plot_triangle(ID)
